@@ -4,7 +4,8 @@
 -behaviour(application).
 -export([start_cli/0,
   string_to_tokens/5, tokens_to_quoted/3, 'string_to_quoted!'/5,
-  env_for_eval/1, quoted_to_erl/2, eval_forms/3, eval_quoted/3]).
+  env_for_eval/1, quoted_to_erl/2, eval_forms/3, eval_quoted/3,
+  eval_quoted/4]).
 -include("elixir.hrl").
 -define(system, 'Elixir.System').
 
@@ -45,7 +46,9 @@ start(_Type, _Args) ->
     {{uri, <<"tftp">>}, 69},
     {{uri, <<"http">>}, 80},
     {{uri, <<"https">>}, 443},
-    {{uri, <<"ldap">>}, 389}
+    {{uri, <<"ldap">>}, 389},
+    {{uri, <<"ws">>}, 80},
+    {{uri, <<"wss">>}, 443}
   ],
 
   Config = [
@@ -56,6 +59,7 @@ start(_Type, _Args) ->
 
     %% Compiler options
     {docs, true},
+    {ignore_already_consolidated, false},
     {ignore_module_conflict, false},
     {parser_options, []},
     {debug_info, true},
@@ -108,10 +112,10 @@ preload_common_modules() ->
 parse_otp_release() ->
   %% Whenever we change this check, we should also change Makefile.
   case string:to_integer(erlang:system_info(otp_release)) of
-    {Num, _} when Num >= 22 ->
+    {Num, _} when Num >= 23 ->
       Num;
     _ ->
-      io:format(standard_error, "ERROR! Unsupported Erlang/OTP version, expected Erlang/OTP 22+~n", []),
+      io:format(standard_error, "ERROR! Unsupported Erlang/OTP version, expected Erlang/OTP 23+~n", []),
       erlang:halt(1)
   end.
 
@@ -170,11 +174,22 @@ env_for_eval(#{lexical_tracker := Pid} = Env) ->
     versioned_vars := #{}
   },
 
-  case is_pid(Pid) andalso is_process_alive(Pid) of
-    true -> NewEnv;
-    false -> NewEnv#{lexical_tracker := nil, tracers := []}
+  case is_pid(Pid) of
+    true ->
+      case is_process_alive(Pid) of
+        true ->
+          NewEnv;
+        false ->
+          'Elixir.IO':warn(
+            <<"an __ENV__ with outdated compilation information was given to eval, "
+              "call Macro.Env.prune_compile_info/1 to prune it">>
+          ),
+          NewEnv#{lexical_tracker := nil, tracers := []}
+      end;
+    false ->
+      NewEnv#{tracers := []}
   end;
-%% TODO: Deprecate all options except line and file.
+%% TODO: Deprecate all options except line, file, module, and function on v1.15.
 env_for_eval(Opts) when is_list(Opts) ->
   Env = elixir_env:new(),
 
@@ -246,22 +261,27 @@ env_for_eval(Opts) when is_list(Opts) ->
 
 %% Quoted evaluation
 
-eval_quoted(Tree, Binding, Opts) when is_list(Opts) ->
-  eval_quoted(Tree, Binding, env_for_eval(Opts));
-eval_quoted(Tree, Binding, #{line := Line} = E) ->
-  eval_forms(elixir_quote:linify(Line, line, Tree), Binding, E).
+eval_quoted(Tree, Binding, E) ->
+  eval_quoted(Tree, Binding, E, []).
+eval_quoted(Tree, Binding, #{line := Line} = E, Opts) ->
+  eval_forms(elixir_quote:linify(Line, line, Tree), Binding, E, Opts).
 
-eval_forms(Tree, Binding, Opts) when is_list(Opts) ->
-  eval_forms(Tree, Binding, env_for_eval(Opts));
-eval_forms(Tree, RawBinding, OrigE) ->
-  {Vars, Binding} = normalize_binding(RawBinding, [], [], 0),
-  E = elixir_env:with_vars(OrigE, Vars),
-  {_, S} = elixir_env:env_to_erl(E),
-  {Erl, NewErlS, NewExS, NewE} = quoted_to_erl(Tree, E, S),
+eval_forms(Tree, Binding, OrigE) ->
+  eval_forms(Tree, Binding, OrigE, []).
+eval_forms(Tree, Binding, OrigE, Opts) ->
+  Prune = proplists:get_value(prune_binding, Opts, false),
+  {ExVars, ErlVars, ErlBinding} = elixir_erl_var:load_binding(Binding, Prune),
+  E = elixir_env:with_vars(OrigE, ExVars),
+  ExS = elixir_env:env_to_ex(E),
+  ErlS = elixir_erl_var:from_env(E, ErlVars),
+  {Erl, NewErlS, NewExS, NewE} = quoted_to_erl(Tree, ErlS, ExS, E),
 
   case Erl of
-    {atom, _, Atom} ->
-      {Atom, RawBinding, NewE};
+    {Literal, _, Value} when Literal == atom; Literal == float; Literal == integer ->
+      if
+        Prune -> {Value, [], NewE#{versioned_vars := #{}}};
+        true -> {Value, Binding, NewE}
+      end;
 
     _  ->
       Exprs =
@@ -270,86 +290,102 @@ eval_forms(Tree, RawBinding, OrigE) ->
           _ -> [Erl]
         end,
 
-      ErlBinding = elixir_erl_var:load_binding(Binding, E, S),
-      {value, Value, NewBinding} = recur_eval(Exprs, ErlBinding, NewE),
-      {Value, elixir_erl_var:dump_binding(NewBinding, NewExS, NewErlS), NewE}
+      ExternalHandler = eval_external_handler(NewE),
+      {value, Value, NewBinding} = erl_eval:exprs(Exprs, ErlBinding, none, ExternalHandler),
+      PruneBefore = if Prune -> length(Binding); true -> -1 end,
+
+      {DumpedBinding, DumpedVars} =
+        elixir_erl_var:dump_binding(NewBinding, NewErlS, NewExS, PruneBefore),
+
+      {Value, DumpedBinding, NewE#{versioned_vars := DumpedVars}}
   end.
 
-normalize_binding([{Key, Value} | Binding], Vars, Acc, I) when is_atom(Key) ->
-  normalize_binding(Binding, [{{Key, nil}, I} | Vars], [{{Key, nil}, Value} | Acc], I + 1);
-normalize_binding([{Pair, Value} | Binding], Vars, Acc, I) ->
-  normalize_binding(Binding, [{Pair, I} | Vars], [{Pair, Value} | Acc], I + 1);
-normalize_binding([], Vars, Acc, _I) ->
-  {maps:from_list(Vars), Acc}.
-
-recur_eval([Expr | Exprs], Binding, Env) ->
-  {value, Value, NewBinding} =
-    % Below must be all one line for locations to be the same
-    % when the stacktrace is extended to the full stacktrace.
-    try erl_eval:expr(Expr, Binding, none, none, none) catch Class:Exception:Stacktrace -> erlang:raise(Class, rewrite_exception(Exception, Stacktrace, Expr, Env), rewrite_stacktrace(Stacktrace)) end,
-
-  case Exprs of
-    [] -> {value, Value, NewBinding};
-    _ -> recur_eval(Exprs, NewBinding, Env)
-  end.
-
-rewrite_exception(badarg, [{Mod, _, _, _} | _], Erl, #{file := File}) when Mod == erl_eval; Mod == eval_bits ->
-  {Min, Max} =
-    erl_parse:fold_anno(fun(Anno, {Min, Max}) ->
-      case erl_anno:line(Anno) of
-        Line when Line > 0 -> {min(Min, Line), max(Max, Line)};
-        _ -> {Min, Max}
+%% TODO: Remove conditional once we require Erlang/OTP 25+.
+-if(?OTP_RELEASE >= 25).
+eval_external_handler(Env) ->
+  Fun = fun(Ann, FunOrModFun, Args) ->
+    try
+      case FunOrModFun of
+          {Mod, Fun} -> apply(Mod, Fun, Args);
+          Fun -> apply(Fun, Args)
       end
-    end, {999999, -999999}, Erl),
+    catch
+      Kind:Reason:Stacktrace ->
+        %% Take everything up to the Elixir module
+        Pruned =
+            lists:takewhile(fun
+              ({elixir,_,_,_}) -> false;
+              (_) -> true
+            end, Stacktrace),
 
-  'Elixir.ArgumentError':exception(
-    erlang:iolist_to_binary(
-      ["argument error while evaluating", badarg_file(File), badarg_line(Min, Max)]
-    )
-  );
-rewrite_exception(Other, _, _, _) ->
-  Other.
+        Caller =
+            lists:dropwhile(fun
+              ({elixir,_,_,_}) -> false;
+              (_) -> true
+            end, Stacktrace),
 
-badarg_file(<<"nofile">>) -> "";
-badarg_file(Path) -> [$\s, 'Elixir.Path':relative_to_cwd(Path)].
+        %% Now we prune any shared code path from erl_eval
+        {current_stacktrace, Current} =
+            erlang:process_info(self(), current_stacktrace),
 
-badarg_line(999999, -999999) -> [];
-badarg_line(Line, Line) -> [" at line ", integer_to_binary(Line)];
-badarg_line(Min, Max) -> [" between lines ", integer_to_binary(Min), " and ", integer_to_binary(Max)].
+        %% We need to make sure that we don't generate more
+        %% frames than supported. So we do our best to drop
+        %% from the Caller, but if the caller has no frames,
+        %% we need to drop from Pruned.
+        {DroppedCaller, ToDrop} =
+          case Caller of
+            [] -> {[], true};
+            _ -> {lists:droplast(Caller), false}
+          end,
 
-rewrite_stacktrace(Stacktrace) ->
-  % eval_eval and eval_bits can call :erlang.raise/3 without the full
-  % stacktrace. When this occurs re-add the current stacktrace so that no
-  % stack information is lost.
-  {current_stacktrace, CurrentStack} = erlang:process_info(self(), current_stacktrace),
-  merge_stacktrace(Stacktrace, tl(CurrentStack)).
+        Reversed = drop_common(lists:reverse(Current), lists:reverse(Pruned), ToDrop),
+        File = elixir_utils:characters_to_list(?key(Env, file)),
+        Location = [{file, File}, {line, erl_anno:line(Ann)}],
 
-% The stacktrace did not include the current stack, re-add it.
-merge_stacktrace([], CurrentStack) ->
-  CurrentStack;
-% The stacktrace includes the current stack.
-merge_stacktrace(CurrentStack, CurrentStack) ->
-  CurrentStack;
-merge_stacktrace([StackItem | Stacktrace], CurrentStack) ->
-  [StackItem | merge_stacktrace(Stacktrace, CurrentStack)].
+        %% Add file+line information at the bottom
+        Custom = lists:reverse([{elixir_eval, '__FILE__', 1, Location} | Reversed], DroppedCaller),
+        erlang:raise(Kind, Reason, Custom)
+    end
+  end,
+  {value, Fun}.
+
+%% We need to check if we have dropped any frames.
+%% If we have not dropped frames, then we need to drop one
+%% at the end so we can put the elixir_eval frame in. If
+%% we have more traces then depth, Erlang would discard
+%% the whole stacktrace.
+drop_common([H | T1], [H | T2], _ToDrop) -> drop_common(T1, T2, false);
+drop_common([_ | T1], T2, ToDrop) -> drop_common(T1, T2, ToDrop);
+drop_common([], [{?MODULE, _, _, _} | T2], _ToDrop) -> T2;
+drop_common([], [_ | T2], true) -> T2;
+drop_common([], T2, _) -> T2.
+-else.
+eval_external_handler(_Env) ->
+  none.
+-endif.
 
 %% Converts a quoted expression to Erlang abstract format
 
 quoted_to_erl(Quoted, E) ->
-  {_, S} = elixir_env:env_to_erl(E),
-  quoted_to_erl(Quoted, E, S).
+  {_, ErlS} = elixir_erl_var:from_env(E),
+  ExS = elixir_env:env_to_ex(E),
+  quoted_to_erl(Quoted, ErlS, ExS, E).
 
-quoted_to_erl(Quoted, Env, Scope) ->
-  {Expanded, NewExS, NewEnv} = elixir_expand:expand(Quoted, elixir_env:env_to_ex(Env), Env),
-  {Erl, NewErlS} = elixir_erl_pass:translate(Expanded, erl_anno:new(?key(Env, line)), Scope),
+quoted_to_erl(Quoted, ErlS, ExS, Env) ->
+  {Expanded, NewExS, NewEnv} =
+    elixir_expand:expand(Quoted, ExS, Env),
+  {Erl, NewErlS} = elixir_erl_pass:translate(Expanded, erl_anno:new(?key(Env, line)), ErlS),
   {Erl, NewErlS, NewExS, NewEnv}.
 
 %% Converts a given string (charlist) into quote expression
 
 string_to_tokens(String, StartLine, StartColumn, File, Opts) when is_integer(StartLine), is_binary(File) ->
-  case elixir_tokenizer:tokenize(String, StartLine, StartColumn, [{file, File} | Opts]) of
+  case elixir_tokenizer:tokenize(String, StartLine, StartColumn, Opts) of
+    {ok, _Line, _Column, [], Tokens} ->
+      {ok, Tokens};
     {ok, _Line, _Column, Warnings, Tokens} ->
-      [elixir_errors:erl_warn(L, F, M) || {L, F, M} <- lists:reverse(Warnings)],
+      (lists:keyfind(emit_warnings, 1, Opts) /= {emit_warnings, false}) andalso
+        [elixir_errors:erl_warn(L, File, M) || {L, M} <- lists:reverse(Warnings)],
       {ok, Tokens};
     {error, {Line, Column, {ErrorPrefix, ErrorSuffix}, Token}, _Rest, _Warnings, _SoFar} ->
       Location = [{line, Line}, {column, Column}],
@@ -404,9 +440,14 @@ parser_location(Meta) ->
   end.
 
 to_binary(List) when is_list(List) -> elixir_utils:characters_to_binary(List);
-to_binary(Atom) when is_atom(Atom) -> atom_to_binary(Atom, utf8).
+to_binary(Atom) when is_atom(Atom) -> atom_to_binary(Atom).
 
-handle_parsing_opts(WarningFile, Opts) ->
+handle_parsing_opts(File, Opts) ->
+  WarningFile =
+    case lists:keyfind(emit_warnings, 1, Opts) of
+      {emit_warnings, false} -> nil;
+      _ -> File
+    end,
   LiteralEncoder =
     case lists:keyfind(literal_encoder, 1, Opts) of
       {literal_encoder, Fun} -> Fun;
