@@ -193,6 +193,28 @@ defmodule MapTest do
              %{a: 1, b: 2, c: :x, d: 5}
   end
 
+  test "replace/3" do
+    map = %{c: 3, b: 2, a: 1}
+    assert Map.replace(map, :b, 10) == %{c: 3, b: 10, a: 1}
+    assert Map.replace(map, :a, 1) == map
+    assert Map.replace(map, :x, 1) == map
+    assert Map.replace(%{}, :x, 1) == %{}
+  end
+
+  test "replace!/3" do
+    map = %{c: 3, b: 2, a: 1}
+    assert Map.replace!(map, :b, 10) == %{c: 3, b: 10, a: 1}
+    assert Map.replace!(map, :a, 1) == map
+
+    assert_raise KeyError, "key :x not found in: %{a: 1, b: 2, c: 3}", fn ->
+      Map.replace!(map, :x, 10)
+    end
+
+    assert_raise KeyError, "key :x not found in: %{}", fn ->
+      Map.replace!(%{}, :x, 10)
+    end
+  end
+
   test "implements (almost) all functions in Keyword" do
     assert Keyword.__info__(:functions) -- Map.__info__(:functions) == [
              delete: 3,
@@ -201,7 +223,9 @@ defmodule MapTest do
              keyword?: 1,
              pop_first: 2,
              pop_first: 3,
-             pop_values: 2
+             pop_values: 2,
+             validate: 2,
+             validate!: 2
            ]
   end
 
@@ -369,6 +393,10 @@ defmodule MapTest do
       defmodule TestMod do
         @enforce_keys :foo
         defstruct [:foo]
+
+        # Verify it remain set afterwards
+        :foo = @enforce_keys
+
         def foo do
           %TestMod{}
         end
