@@ -125,7 +125,7 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warning = """
-      warning: List.old_flatten/1 is undefined or private. Did you mean one of:
+      warning: List.old_flatten/1 is undefined or private. Did you mean:
 
             * flatten/1
             * flatten/2
@@ -137,39 +137,6 @@ defmodule Module.Types.IntegrationTest do
       assert_warnings(files, warning)
     end
 
-    test "aliases" do
-      files = %{
-        "a.ex" => """
-        defmodule A do
-          alias Enum, as: E
-
-          def a(a, b), do: E.map2(a, b)
-          def b, do: &E.map2/2
-
-          @file "external_source.ex"
-          def c do
-            alias Enum, as: EE
-            &EE.map2/2
-          end
-        end
-        """
-      }
-
-      warning = """
-      warning: Enum.map2/2 is undefined or private. Did you mean one of:
-
-            * map/2
-
-      Found at 3 locations:
-        a.ex:4: A.a/2
-        a.ex:5: A.b/0
-        external_source.ex:10: A.c/0
-
-      """
-
-      assert_warnings(files, warning)
-    end
-
     test "reports missing functions" do
       files = %{
         "a.ex" => """
@@ -209,13 +176,13 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warning = """
-      warning: A.a/1 is undefined or private. Did you mean one of:
+      warning: A.a/1 is undefined or private. Did you mean:
 
             * a/0
 
         a.ex:3: A.b/0
 
-      warning: A.b/1 is undefined or private. Did you mean one of:
+      warning: A.b/1 is undefined or private. Did you mean:
 
             * b/0
 
@@ -234,6 +201,8 @@ defmodule Module.Types.IntegrationTest do
 
           @file "external_source.ex"
           def c, do: E.no_module()
+
+          def i, do: Io.puts "hello"
         end
         """
       }
@@ -245,6 +214,9 @@ defmodule Module.Types.IntegrationTest do
       warning: E.no_module/0 is undefined (module E is not available or is yet to be defined)
         external_source.ex:5: A.c/0
 
+      warning: Io.puts/1 is undefined (module Io is not available or is yet to be defined)
+        a.ex:7: A.i/0
+
       """
 
       assert_warnings(files, warning)
@@ -277,13 +249,13 @@ defmodule Module.Types.IntegrationTest do
     test "doesn't report missing funcs at compile time" do
       files = %{
         "a.ex" => """
-          Enum.map([], fn _ -> BadReferencer.no_func4() end)
+        Enum.map([], fn _ -> BadReferencer.no_func4() end)
 
-          if function_exported?(List, :flatten, 1) do
-            List.flatten([1, 2, 3])
-          else
-            List.old_flatten([1, 2, 3])
-          end
+        if function_exported?(List, :flatten, 1) do
+          List.flatten([1, 2, 3])
+        else
+          List.old_flatten([1, 2, 3])
+        end
         """
       }
 
@@ -337,12 +309,12 @@ defmodule Module.Types.IntegrationTest do
 
       warning = """
       warning: A2.no_func/0 is undefined (module A2 is not available or is yet to be defined)
-      Found at 2 locations:
+      Invalid call found at 2 locations:
         a.ex:8: A.d/0
         external_source.ex:5: A.b/0
 
       warning: A.no_func/0 is undefined or private
-      Found at 2 locations:
+      Invalid call found at 2 locations:
         a.ex:2: A.a/0
         a.ex:7: A.c/0
 
@@ -403,7 +375,7 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warning = """
-      warning: A.to_list/1 is undefined or private. Did you mean one of:
+      warning: A.to_list/1 is undefined or private. Did you mean:
 
             * to_charlist/1
 
@@ -414,35 +386,6 @@ defmodule Module.Types.IntegrationTest do
       assert_warnings(files, warning)
     end
 
-    test "imports" do
-      files = %{
-        "a.ex" => """
-        defmodule A do
-          import Record
-
-          def a(a, b), do: extract(a, b)
-          def b(arg), do: is_record(arg)
-        end
-        """
-      }
-
-      assert_no_warnings(files)
-    end
-
-    test "requires" do
-      files = %{
-        "a.ex" => """
-        defmodule A do
-          require Integer
-
-          def a(a), do: Integer.is_even(a)
-        end
-        """
-      }
-
-      assert_no_warnings(files)
-    end
-
     test "do not warn for module defined in local context" do
       files = %{
         "a.ex" => """
@@ -513,7 +456,7 @@ defmodule Module.Types.IntegrationTest do
       warning: MissingModule3.func/2 is undefined (module MissingModule3 is not available or is yet to be defined)
         a.ex:8: A.d/0
 
-      warning: B.func/3 is undefined or private. Did you mean one of:
+      warning: B.func/3 is undefined or private. Did you mean:
 
             * func/1
 
@@ -561,7 +504,7 @@ defmodule Module.Types.IntegrationTest do
         warning: MissingModule3.func/2 is undefined (module MissingModule3 is not available or is yet to be defined)
           a.ex:8: A.d/0
 
-        warning: B.func/3 is undefined or private. Did you mean one of:
+        warning: B.func/3 is undefined or private. Did you mean:
 
               * func/1
 
@@ -619,6 +562,45 @@ defmodule Module.Types.IntegrationTest do
     end
   end
 
+  describe "regressions" do
+    test "handle missing location info from quoted" do
+      assert capture_io(:stderr, fn ->
+               quote do
+                 defmodule X do
+                   def f() do
+                     x = %{}
+                     %{x | key: :value}
+                   end
+                 end
+               end
+               |> Code.compile_quoted()
+             end) =~ "warning:"
+    end
+  end
+
+  describe "after_verify" do
+    test "reports functions" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          @after_verify __MODULE__
+
+          def __after_verify__(__MODULE__) do
+            IO.warn "from after_verify", []
+          end
+        end
+        """
+      }
+
+      warning = """
+      warning: from after_verify
+
+      """
+
+      assert_warnings(files, warning)
+    end
+  end
+
   describe "deprecated" do
     test "reports functions" do
       files = %{
@@ -685,13 +667,13 @@ defmodule Module.Types.IntegrationTest do
       warnings = [
         """
         warning: A.__struct__/0 is deprecated. oops
-        Found at 2 locations:
+        Invalid call found at 2 locations:
           a.ex:4: A.match/1
           a.ex:5: A.build/1
         """,
         """
         warning: A.__struct__/0 is deprecated. oops
-        Found at 2 locations:
+        Invalid call found at 2 locations:
           b.ex:2: B.match/1
           b.ex:3: B.build/1
 
@@ -725,6 +707,31 @@ defmodule Module.Types.IntegrationTest do
 
       assert_warnings(files, warning)
     end
+
+    test "reports macro" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          @deprecated "oops"
+          defmacro a, do: :ok
+        end
+        """,
+        "b.ex" => """
+        defmodule B do
+          require A
+          def b, do: A.a()
+        end
+        """
+      }
+
+      warning = """
+      warning: A.a/0 is deprecated. oops
+        b.ex:3: B.b/0
+
+      """
+
+      assert_warnings(files, warning)
+    end
   end
 
   defp assert_warnings(files, expected) when is_binary(expected) do
