@@ -38,6 +38,21 @@ defmodule PathTest do
     File.rm_rf(Path.join(config.tmp_dir, "wildcard"))
   end
 
+  @tag :tmp_dir
+  test "wildcard/2 follows ..", config do
+    hello = Path.join(config.tmp_dir, "wildcard/hello")
+    world = Path.join(config.tmp_dir, "wildcard/world")
+    File.mkdir_p(hello)
+    File.touch(world)
+
+    assert Path.wildcard(Path.join(config.tmp_dir, "wildcard/w*/../h*")) == []
+
+    assert Path.wildcard(Path.join(config.tmp_dir, "wildcard/h*/../w*")) ==
+             [Path.join(config.tmp_dir, "wildcard/hello/../world")]
+  after
+    File.rm_rf(Path.join(config.tmp_dir, "wildcard"))
+  end
+
   test "wildcard/2 raises on null byte" do
     assert_raise ArgumentError, ~r/null byte/, fn -> Path.wildcard("foo\0bar") end
   end
@@ -50,6 +65,16 @@ defmodule PathTest do
       assert Path.absname("\\\\host\\path") == "//host/path"
       assert Path.absname("\\/host\\path") == "//host/path"
       assert Path.absname("/\\host\\path") == "//host/path"
+
+      assert Path.absname("c:/") == "c:/"
+      assert Path.absname("c:/host/path") == "c:/host/path"
+
+      cwd = File.cwd!()
+      assert Path.absname(cwd |> String.split("/") |> hd()) == cwd
+
+      <<letter, _::binary>> = cwd
+      random = Enum.random(Enum.to_list(?c..?z) -- [letter])
+      assert Path.absname(<<random, ?:>>) == <<random, ?:, ?/>>
     end
 
     test "relative/1" do
@@ -91,6 +116,21 @@ defmodule PathTest do
       assert Path.split("C:\\foo\\bar") == ["c:/", "foo", "bar"]
       assert Path.split("C:/foo/bar") == ["c:/", "foo", "bar"]
     end
+
+    test "safe_relative/1" do
+      assert Path.safe_relative("local/foo") == {:ok, "local/foo"}
+      assert Path.safe_relative("D:/usr/local/foo") == :error
+      assert Path.safe_relative("d:/usr/local/foo") == :error
+      assert Path.safe_relative("foo/../..") == :error
+    end
+
+    test "safe_relative_to/2" do
+      assert Path.safe_relative_to("local/foo/bar", "local") == {:ok, "local/foo/bar"}
+      assert Path.safe_relative_to("foo/..", "local") == {:ok, ""}
+      assert Path.safe_relative_to("..", "local/foo") == :error
+      assert Path.safe_relative_to("d:/usr/local/foo", "D:/") == :error
+      assert Path.safe_relative_to("D:/usr/local/foo", "d:/") == :error
+    end
   end
 
   describe "Unix" do
@@ -193,7 +233,9 @@ defmodule PathTest do
     assert Path.relative_to("/usr/local/foo", "/usr/local") == "foo"
     assert Path.relative_to("/usr/local/foo", "/") == "usr/local/foo"
     assert Path.relative_to("/usr/local/foo", "/etc") == "/usr/local/foo"
-    assert Path.relative_to("/usr/local/foo", "/usr/local/foo") == "/usr/local/foo"
+    assert Path.relative_to("/usr/local/foo", "/usr/local/foo") == "."
+    assert Path.relative_to("/usr/local/foo/", "/usr/local/foo") == "."
+    assert Path.relative_to("/usr/local/foo", "/usr/local/foo/") == "."
 
     assert Path.relative_to("usr/local/foo", "usr/local") == "foo"
     assert Path.relative_to("usr/local/foo", "etc") == "usr/local/foo"
@@ -203,6 +245,22 @@ defmodule PathTest do
     assert Path.relative_to(["usr", ?/, 'local/foo'], 'usr/local') == "foo"
   end
 
+  test "safe_relative/1" do
+    assert Path.safe_relative("foo/bar") == {:ok, "foo/bar"}
+    assert Path.safe_relative("foo/..") == {:ok, ""}
+    assert Path.safe_relative("./foo") == {:ok, "foo"}
+
+    assert Path.safe_relative("/usr/local/foo") == :error
+    assert Path.safe_relative("foo/../..") == :error
+  end
+
+  test "safe_relative_to/2" do
+    assert Path.safe_relative_to("/usr/local/foo", "/usr/local") == :error
+    assert Path.safe_relative_to("../../..", "foo/bar") == :error
+    assert Path.safe_relative_to("../../..", "foo/bar") == :error
+    assert Path.safe_relative_to("/usr/local/foo", "/") == :error
+  end
+
   test "rootname/2" do
     assert Path.rootname("~/foo/bar.ex", ".ex") == "~/foo/bar"
     assert Path.rootname("~/foo/bar.exs", ".ex") == "~/foo/bar.exs"
