@@ -101,29 +101,21 @@ defmodule ExUnit.Assertions do
       assert nil = some_function_that_returns_nil()
 
   Even though the match works, `assert` still expects a truth
-  value. In such cases, simply use `Kernel.==/2` or
-  `Kernel.match?/2`.
+  value. In such cases, simply use `==/2` or `match?/2`.
   """
   defmacro assert({:=, meta, [left, right]} = assertion) do
     code = escape_quoted(:assert, meta, assertion)
 
-    # If the match works, we need to check if the value
-    # is not nil nor false. We need to rewrite the if
-    # to avoid silly warnings though.
     check =
-      suppress_warning(
-        quote do
-          case right do
-            x when x in [nil, false] ->
-              raise ExUnit.AssertionError,
-                expr: expr,
-                message: "Expected truthy, got #{inspect(right)}"
-
-            _ ->
-              :ok
-          end
+      quote generated: true do
+        if right do
+          :ok
+        else
+          raise ExUnit.AssertionError,
+            expr: expr,
+            message: "Expected truthy, got #{inspect(right)}"
         end
-      )
+      end
 
     __match__(left, right, code, check, __CALLER__)
   end
@@ -155,17 +147,15 @@ defmodule ExUnit.Assertions do
     else
       {args, value} = extract_args(assertion, __CALLER__)
 
-      quote do
-        value = unquote(value)
-
-        unless value do
+      quote generated: true do
+        if value = unquote(value) do
+          value
+        else
           raise ExUnit.AssertionError,
             args: unquote(args),
             expr: unquote(escape_quoted(:assert, [], assertion)),
             message: "Expected truthy, got #{inspect(value)}"
         end
-
-        value
       end
     end
   end
@@ -181,8 +171,7 @@ defmodule ExUnit.Assertions do
   The code above will fail because the `=` operator always fails
   when the sides do not match and `refute/2` does not change it.
 
-  The correct way to write the refutation above is to use
-  `Kernel.match?/2`:
+  The correct way to write the refutation above is to use `match?/2`:
 
       refute match?({:ok, _}, some_function_that_returns_error_tuple())
 
@@ -219,17 +208,15 @@ defmodule ExUnit.Assertions do
     else
       {args, value} = extract_args(assertion, __CALLER__)
 
-      quote do
-        value = unquote(value)
-
-        if value do
+      quote generated: true do
+        if value = unquote(value) do
           raise ExUnit.AssertionError,
             args: unquote(args),
             expr: unquote(escape_quoted(:refute, [], assertion)),
             message: "Expected false or nil, got #{inspect(value)}"
+        else
+          value
         end
-
-        value
       end
     end
   end
@@ -240,22 +227,26 @@ defmodule ExUnit.Assertions do
 
   defp translate_assertion(:assert, {operator, meta, [_, _]} = expr, caller)
        when operator in @operator do
-    left = Macro.var(:left, __MODULE__)
-    right = Macro.var(:right, __MODULE__)
-    call = {operator, meta, [left, right]}
-    equality_check? = operator in [:<, :>, :!==, :!=]
-    message = "Assertion with #{operator} failed"
-    translate_operator(:assert, expr, call, message, equality_check?, caller)
+    if match?([{_, Kernel}], Macro.Env.lookup_import(caller, {operator, 2})) do
+      left = Macro.var(:left, __MODULE__)
+      right = Macro.var(:right, __MODULE__)
+      call = {operator, meta, [left, right]}
+      equality_check? = operator in [:<, :>, :!==, :!=]
+      message = "Assertion with #{operator} failed"
+      translate_operator(:assert, expr, call, message, equality_check?, caller)
+    end
   end
 
   defp translate_assertion(:refute, {operator, meta, [_, _]} = expr, caller)
        when operator in @operator do
-    left = Macro.var(:left, __MODULE__)
-    right = Macro.var(:right, __MODULE__)
-    call = {:not, meta, [{operator, meta, [left, right]}]}
-    equality_check? = operator in [:<=, :>=, :===, :==, :=~]
-    message = "Refute with #{operator} failed"
-    translate_operator(:refute, expr, call, message, equality_check?, caller)
+    if match?([{_, Kernel}], Macro.Env.lookup_import(caller, {operator, 2})) do
+      left = Macro.var(:left, __MODULE__)
+      right = Macro.var(:right, __MODULE__)
+      call = {:not, meta, [{operator, meta, [left, right]}]}
+      equality_check? = operator in [:<=, :>=, :===, :==, :=~]
+      message = "Refute with #{operator} failed"
+      translate_operator(:refute, expr, call, message, equality_check?, caller)
+    end
   end
 
   defp translate_assertion(_kind, _expected, _caller) do
@@ -269,19 +260,21 @@ defmodule ExUnit.Assertions do
     quote do
       left = unquote(left)
       right = unquote(right)
+      expr = unquote(expr)
+      message = unquote(message)
 
       if ExUnit.Assertions.__equal__?(left, right) do
         ExUnit.Assertions.assert(false,
           left: left,
-          expr: unquote(expr),
-          message: unquote(message <> ", both sides are exactly equal")
+          expr: expr,
+          message: message <> ", both sides are exactly equal"
         )
       else
         ExUnit.Assertions.assert(unquote(call),
           left: left,
           right: right,
-          expr: unquote(expr),
-          message: unquote(message),
+          expr: expr,
+          message: message,
           context: unquote(context)
         )
       end
@@ -388,6 +381,10 @@ defmodule ExUnit.Assertions do
 
       assert false, "it will never be true"
 
+      assert x == :foo, "expected x to be foo"
+
+      assert match?({:ok, _}, x), "expected x to match {:ok, _}"
+
   """
   def assert(value, message) when is_binary(message) do
     assert(value, message: message)
@@ -692,6 +689,9 @@ defmodule ExUnit.Assertions do
       {:__aliases__, _, _} = expr ->
         Macro.expand(expr, caller)
 
+      {:@, _, [{attribute, _, _}]} ->
+        caller.module |> Module.get_attribute(attribute) |> Macro.escape()
+
       {left, meta, right} = expr ->
         case Macro.expand(expr, caller) do
           ^expr -> expr
@@ -850,7 +850,7 @@ defmodule ExUnit.Assertions do
       assert catch_exit(exit(1)) == 1
 
   To assert exits from linked processes started from the test, trap exits
-  with `Process.flag/2` and assert the exit message with `assert_received/2`.
+  with `Process.flag/2` and assert the exit message with `assert_receive/2`.
 
       Process.flag(:trap_exit, true)
       pid = spawn_link(fn -> Process.exit(self(), :normal) end)
