@@ -14,8 +14,8 @@ defmodule ExUnit.DocTest do
   Multiline expressions can be used by prefixing subsequent lines
   with either `...>` (recommended) or `iex>`.
 
-  The expected result should start at the next line after the `iex>`
-  or `...>` line(s) and it is terminated either by a newline.
+  The expected result should start the line after the `iex>`
+  and `...>` line(s) and be terminated by a newline.
 
   ## Examples
 
@@ -131,7 +131,7 @@ defmodule ExUnit.DocTest do
       iex(1)> raise "some error"
       ** (RuntimeError) some error
 
-  Doctest will looking for a line starting with `** (` and it will parse it
+  Doctest will look for a line starting with `** (` and it will parse it
   accordingly to extract the exception name and message. The exception parser
   will consider all following lines part of the exception message until there
   is an empty line or there is a new expression prefixed with `iex>`.
@@ -159,7 +159,7 @@ defmodule ExUnit.DocTest do
       module = Keyword.fetch!(opts, :module)
       message = Keyword.fetch!(opts, :message)
 
-      file = module.__info__(:compile)[:source] |> Path.relative_to_cwd()
+      file = module.module_info(:compile)[:source] |> Path.relative_to_cwd()
       info = Exception.format_file_line(file, opts[:line])
       %__MODULE__{message: info <> " " <> message}
     end
@@ -182,8 +182,7 @@ defmodule ExUnit.DocTest do
     * `:import` - when `true`, one can test a function defined in the module
       without referring to the module name. However, this is not feasible when
       there is a clash with a module like `Kernel`. In these cases, `:import`
-      should be set to `false` and a full `Module.function` construct should be
-      used.
+      should be set to `false` and `Module.function(...)` should be used instead.
 
     * `:tags` - a list of tags to apply to all generated doctests.
 
@@ -194,25 +193,27 @@ defmodule ExUnit.DocTest do
   This macro is auto-imported with every `ExUnit.Case`.
   """
   defmacro doctest(module, opts \\ []) do
+    caller = __CALLER__
+
     require =
-      if is_atom(Macro.expand(module, __CALLER__)) do
+      if is_atom(Macro.expand(module, caller)) do
         quote do
           require unquote(module)
         end
       end
 
     tests =
-      quote bind_quoted: [module: module, opts: opts] do
-        env = __ENV__
+      quote bind_quoted: [
+              module: module,
+              opts: opts,
+              env_line: caller.line,
+              env_file: caller.file
+            ] do
         file = ExUnit.DocTest.__file__(module)
 
-        for {name, test} <- ExUnit.DocTest.__doctests__(module, opts) do
-          if tags = Keyword.get(opts, :tags) do
-            @tag tags
-          end
-
+        for {name, test, tags} <- ExUnit.DocTest.__doctests__(module, opts) do
           @file file
-          doc = ExUnit.Case.register_test(env, :doctest, name, [])
+          doc = ExUnit.Case.register_test(__MODULE__, env_file, env_line, :doctest, name, tags)
           def unquote(doc)(_), do: unquote(test)
         end
       end
@@ -223,7 +224,7 @@ defmodule ExUnit.DocTest do
   @doc false
   def __file__(module) do
     source =
-      module.__info__(:compile)[:source] ||
+      module.module_info(:compile)[:source] ||
         raise "#{inspect(module)} does not have compile-time source information"
 
     "(for doctest at) " <> Path.relative_to_cwd(source)
@@ -231,37 +232,63 @@ defmodule ExUnit.DocTest do
 
   @doc false
   def __doctests__(module, opts) do
-    do_import = Keyword.get(opts, :import, false)
+    tags = [doctest: module] ++ Keyword.get(opts, :tags, [])
+    import = Keyword.get(opts, :import, false)
+    file = module.module_info(:compile)[:source] |> Path.relative_to_cwd()
 
     extract(module)
-    |> filter_by_opts(opts)
-    |> Stream.with_index()
-    |> Enum.map(fn {test, acc} ->
-      compile_test(test, module, do_import, acc + 1)
+    |> filter_by_opts(module, opts)
+    |> Enum.sort_by(& &1.line)
+    |> Enum.with_index(fn test, index ->
+      compile_test(test, module, import, index + 1, file, tags)
     end)
   end
 
-  defp filter_by_opts(tests, opts) do
+  defp filter_by_opts(tests, module, opts) do
     except = Keyword.get(opts, :except, [])
 
     case Keyword.fetch(opts, :only) do
-      {:ok, []} ->
-        []
-
-      {:ok, only} ->
-        tests
-        |> Stream.reject(&(&1.fun_arity in except))
-        |> Stream.filter(&(&1.fun_arity in only))
-
-      :error ->
-        Stream.reject(tests, &(&1.fun_arity in except))
+      {:ok, []} -> []
+      {:ok, only} -> filter_tests(module, tests, except, only)
+      :error -> Stream.reject(tests, &(&1.fun_arity in except))
     end
   end
 
+  defp filter_tests(module, tests, except, only) do
+    {filtered_tests, fun_arities} =
+      for test <- tests,
+          test.fun_arity not in except,
+          test.fun_arity in only,
+          reduce: {[], []} do
+        {tests, fun_arities} -> {[test | tests], [test.fun_arity | fun_arities]}
+      end
+
+    case only -- [:moduledoc | fun_arities] do
+      [] ->
+        filtered_tests
+
+      undefined_fun_arities ->
+        pluralized = pluralize_list_name("function", undefined_fun_arities)
+
+        functions =
+          Enum.map_join(undefined_fun_arities, "\n    ", fn {fun, arity} ->
+            Exception.format_mfa(module, fun, arity)
+          end)
+
+        raise Error,
+          module: module,
+          message: "undefined or private #{pluralized} given to doctest:\n\n    #{functions}\n\n"
+    end
+  end
+
+  defp pluralize_list_name(name, [_]), do: name
+  defp pluralize_list_name(name, _), do: ExUnit.plural_rule(name)
+
   ## Compilation of extracted tests
 
-  defp compile_test(test, module, do_import, n) do
-    {test_name(test, module, n), test_content(test, module, do_import)}
+  defp compile_test(test, module, do_import, n, file, tags) do
+    tags = [doctest_line: test.line] ++ tags
+    {test_name(test, module, n), test_content(test, module, do_import, file), tags}
   end
 
   defp test_name(%{fun_arity: :moduledoc}, m, n) do
@@ -272,8 +299,7 @@ defmodule ExUnit.DocTest do
     "#{inspect(m)}.#{f}/#{a} (#{n})"
   end
 
-  defp test_content(%{exprs: exprs, line: line}, module, do_import) do
-    file = module.__info__(:compile)[:source] |> Path.relative_to_cwd()
+  defp test_content(%{exprs: exprs, line: line}, module, do_import, file) do
     location = [line: line, file: file]
     stack = Macro.escape([{module, :__MODULE__, 0, location}])
 
@@ -281,12 +307,14 @@ defmodule ExUnit.DocTest do
       raise Error,
         line: line,
         module: module,
-        message: "multiple exceptions in one doctest case are not supported"
+        message:
+          "multiple exceptions in the same doctest example are not supported, " <>
+            "please separate your iex> prompts by multiple newlines to start new examples"
     end
 
     tests =
-      Enum.map(exprs, fn {expr, expected, formatted} ->
-        test_case_content(expr, expected, location, stack, formatted)
+      Enum.map(exprs, fn {expr, expected, doctest} ->
+        test_case_content(expr, expected, location, stack, doctest)
       end)
 
     {:__block__, [], test_import(module, do_import) ++ tests}
@@ -299,119 +327,132 @@ defmodule ExUnit.DocTest do
     end) > 1
   end
 
-  defp test_case_content(expr, :test, location, stack, formatted) do
-    string_to_quoted(location, stack, expr, "\n" <> formatted) |> insert_assertions()
+  defp test_case_content(expr, :test, location, stack, doctest) do
+    string_to_quoted(location, stack, expr, doctest) |> insert_assertions()
   end
 
-  defp test_case_content(expr, {:test, expected}, location, stack, formatted) do
-    doctest = "\n" <> formatted <> "\n" <> expected
+  defp test_case_content(expr, {:test, expected}, location, stack, doctest) do
     expr_ast = string_to_quoted(location, stack, expr, doctest) |> insert_assertions()
     expected_ast = string_to_quoted(location, stack, expected, doctest)
-    last_expr_ast = last_expr(expr_ast)
+    last_expr = Macro.to_string(last_expr(expr_ast))
 
     quote do
+      value = unquote(expr_ast)
       expected = unquote(expected_ast)
+      doctest = unquote(doctest)
+      last_expr = unquote(last_expr)
+      expected_expr = unquote(expected)
+      stack = unquote(stack)
 
-      case unquote(expr_ast) do
-        ^expected ->
-          :ok
-
-        actual ->
-          doctest = unquote(doctest)
-
-          expr =
-            "#{unquote(Macro.to_string(last_expr_ast))} === #{unquote(String.trim(expected))}"
-
-          error = [
-            message: "Doctest failed",
-            doctest: doctest,
-            expr: expr,
-            left: actual,
-            right: expected
-          ]
-
-          reraise ExUnit.AssertionError, error, unquote(stack)
-      end
+      ExUnit.DocTest.__test__(value, expected, doctest, last_expr, expected_expr, stack)
     end
   end
 
-  defp test_case_content(expr, {:inspect, expected}, location, stack, formatted) do
-    doctest = "\n" <> formatted <> "\n" <> expected
+  defp test_case_content(expr, {:inspect, expected}, location, stack, doctest) do
     expr_ast = string_to_quoted(location, stack, expr, doctest) |> insert_assertions()
-    expected_ast = string_to_quoted(location, stack, expected, doctest)
-    last_expr_ast = last_expr(expr_ast)
+    last_expr = Macro.to_string(last_expr(expr_ast))
 
     quote do
-      expected = unquote(expected_ast)
+      value = unquote(expr_ast)
+      expected = unquote(expected)
+      doctest = unquote(doctest)
+      last_expr = unquote(last_expr)
+      expected_expr = unquote(inspect(expected))
+      stack = unquote(stack)
 
-      case inspect(unquote(expr_ast)) do
-        ^expected ->
-          :ok
-
-        actual ->
-          doctest = unquote(doctest)
-
-          expr =
-            "inspect(#{unquote(Macro.to_string(last_expr_ast))}) === " <>
-              "#{unquote(String.trim(expected))}"
-
-          error = [
-            message: "Doctest failed",
-            doctest: doctest,
-            expr: expr,
-            left: actual,
-            right: expected
-          ]
-
-          reraise ExUnit.AssertionError, error, unquote(stack)
-      end
+      ExUnit.DocTest.__inspect__(value, expected, doctest, last_expr, expected_expr, stack)
     end
   end
 
-  defp test_case_content(expr, {:error, exception, message}, location, stack, formatted) do
-    doctest = "\n" <> formatted <> "\n** (#{inspect(exception)}) #{inspect(message)}"
+  defp test_case_content(expr, {:error, exception, message}, location, stack, doctest) do
     expr_ast = string_to_quoted(location, stack, expr, doctest)
 
     quote do
       stack = unquote(stack)
-      exception = unquote(exception)
-      doctest = unquote(doctest)
       message = unquote(message)
+      doctest = unquote(doctest)
+      exception = unquote(exception)
+      ExUnit.DocTest.__error__(fn -> unquote(expr_ast) end, message, exception, doctest, stack)
+    end
+  end
 
+  @doc false
+  def __test__(value, expected, doctest, last_expr, expected_expr, stack) do
+    case value do
+      ^expected ->
+        {:ok, value}
+
+      _ ->
+        error = [
+          message: "Doctest failed",
+          doctest: doctest,
+          expr: "#{last_expr} === #{String.trim(expected_expr)}",
+          left: value,
+          right: expected
+        ]
+
+        reraise ExUnit.AssertionError, error, stack
+    end
+  end
+
+  @doc false
+  def __inspect__(value, expected, doctest, last_expr, expected_expr, parent_stack) do
+    result =
       try do
-        unquote(expr_ast)
+        inspect(value, safe: false)
       rescue
-        error ->
-          actual_exception = error.__struct__
-          actual_message = Exception.message(error)
-
-          message =
-            cond do
-              actual_exception != exception ->
-                "Doctest failed: expected exception #{inspect(exception)} but got " <>
-                  "#{inspect(actual_exception)} with message #{inspect(actual_message)}"
-
-              actual_message != message ->
-                "Doctest failed: wrong message for #{inspect(actual_exception)}\n" <>
-                  "expected:\n" <>
-                  "  #{inspect(message)}\n" <>
-                  "actual:\n" <> "  #{inspect(actual_message)}"
-
-              true ->
-                nil
-            end
-
-          if message do
-            reraise ExUnit.AssertionError, [message: message, doctest: doctest], stack
-          end
+        e ->
+          stack = Enum.drop(__STACKTRACE__, 1)
+          {[message: Exception.message(e)], ExUnit.Runner.prune_stacktrace(stack)}
       else
-        _ ->
-          message =
-            "Doctest failed: expected exception #{inspect(exception)} but nothing was raised"
-
-          error = [message: message, doctest: doctest]
-          reraise ExUnit.AssertionError, error, stack
+        ^expected -> :ok
+        actual -> {[left: actual, right: expected, message: "Doctest failed"], []}
       end
+
+    case result do
+      :ok ->
+        {:ok, value}
+
+      {extra, stack} ->
+        expr = "inspect(#{last_expr}) === #{String.trim(expected_expr)}"
+        error = [doctest: doctest, expr: expr] ++ extra
+        reraise ExUnit.AssertionError, error, stack ++ parent_stack
+    end
+  end
+
+  @doc false
+  def __error__(fun, message, exception, doctest, stack) do
+    try do
+      fun.()
+    rescue
+      error ->
+        actual_exception = error.__struct__
+        actual_message = Exception.message(error)
+
+        failed =
+          cond do
+            actual_exception != exception ->
+              "Doctest failed: expected exception #{inspect(exception)} but got " <>
+                "#{inspect(actual_exception)} with message #{inspect(actual_message)}"
+
+            actual_message != message ->
+              "Doctest failed: wrong message for #{inspect(actual_exception)}\n" <>
+                "expected:\n" <>
+                "  #{inspect(message)}\n" <>
+                "actual:\n" <> "  #{inspect(actual_message)}"
+
+            true ->
+              nil
+          end
+
+        if failed do
+          reraise ExUnit.AssertionError, [message: failed, doctest: doctest], stack
+        end
+    else
+      _ ->
+        failed = "Doctest failed: expected exception #{inspect(exception)} but nothing was raised"
+        error = [message: failed, doctest: doctest]
+        reraise ExUnit.AssertionError, error, stack
     end
   end
 
@@ -430,7 +471,7 @@ defmodule ExUnit.DocTest do
           if e.__struct__ == TokenMissingError and expr =~ @opaque_type_regex do
             message <>
               """
-              . If you are planning to assert on the result of an iex> expression \
+              \nIf you are planning to assert on the result of an iex> expression \
               which contains a value inspected as #Name<...>, please make sure \
               the inspected value is placed at the beginning of the expression; \
               otherwise Elixir will treat it as a comment due to the leading sign #.\
@@ -478,22 +519,18 @@ defmodule ExUnit.DocTest do
   defp explain_docs_error({:invalid_chunk, _}),
     do: "The documentation chunk in the module is invalid"
 
-  defp extract_from_moduledoc(_, doc, _module) when doc in [:none, :hidden], do: []
-
   defp extract_from_moduledoc(annotation, %{"en" => doc}, module) do
     for test <- extract_tests(:erl_anno.line(annotation), doc, module) do
       normalize_test(test, :moduledoc)
     end
   end
 
+  defp extract_from_moduledoc(_, _doc, _module), do: []
+
   defp extract_from_docs(docs, module) do
     for doc <- docs, doc <- extract_from_doc(doc, module), do: doc
   end
 
-  defp extract_from_doc({{kind, _, _}, _, _, doc, _}, _module)
-       when kind not in [:function, :macro, :type] or doc in [:none, :hidden],
-       do: []
-
   defp extract_from_doc({{_, name, arity}, annotation, _, %{"en" => doc}, _}, module) do
     line = :erl_anno.line(annotation)
 
@@ -502,8 +539,11 @@ defmodule ExUnit.DocTest do
     end
   end
 
+  defp extract_from_doc(_doc, _module),
+    do: []
+
   defp extract_tests(line_no, doc, module) do
-    all_lines = String.split(doc, "\n", trim: false)
+    all_lines = String.split(doc, ["\r\n", "\n"], trim: false)
     lines = adjust_indent(all_lines, line_no + 1, module)
     extract_tests(lines, "", "", [], true, module, "")
   end
@@ -815,7 +855,8 @@ defmodule ExUnit.DocTest do
   end
 
   defp add_expr(%{exprs: exprs} = test, expr, expected, formatted) do
-    %{test | exprs: [{expr, tag_expected(expected), formatted} | exprs]}
+    doctest = "\n" <> formatted <> "\n" <> expected
+    %{test | exprs: [{expr, tag_expected(expected), doctest} | exprs]}
   end
 
   defp tag_expected(string) do
@@ -829,7 +870,7 @@ defmodule ExUnit.DocTest do
 
       _ ->
         if inspectable?(string) do
-          {:inspect, inspect(string)}
+          {:inspect, string}
         else
           {:test, string}
         end
