@@ -8,11 +8,11 @@ defmodule Mix.Tasks.Test do
   @recursive true
   @preferred_cli_env :test
 
-  @moduledoc """
+  @moduledoc ~S"""
   Runs the tests for a project.
 
   This task starts the current application, loads up
-  `test/test_helper.exs` and then requires all files matching the
+  `test/test_helper.exs` and then, requires all files matching the
   `test/**/*_test.exs` pattern in parallel.
 
   A list of files and/or directories can be given after the task
@@ -31,52 +31,174 @@ defmodule Mix.Tasks.Test do
       # To run a given test file on my_app from the umbrella root
       mix test apps/my_app/test/some/particular/file_test.exs
 
+  ## Understanding test results
+
+  When you run your test suite, it prints results as they run with
+  a summary at the end, as seen below:
+
+      $ mix test
+      ...
+
+        1) test greets the world (FooTest)
+           test/foo_test.exs:5
+           Assertion with == failed
+           code:  assert Foo.hello() == :world!
+           left:  :world
+           right: :world!
+           stacktrace:
+             test/foo_test.exs:6: (test)
+
+      ........
+
+      Finished in 0.05 seconds (0.00s async, 0.05s sync)
+      1 doctest, 11 tests, 1 failure
+
+      Randomized with seed 646219
+
+  For each test, the test suite will print a dot. Failed tests
+  are printed immediately in the format described in the next
+  section.
+
+  After all tests run, we print the suite summary. The first
+  line contains the total time spent on the suite, followed
+  by how much time was spent on async tests (defined with
+  `use ExUnit.Case, async: true`) vs sync ones:
+
+      Finished in 0.05 seconds (0.00s async, 0.05s sync)
+
+  Developers want to minimize the time spent on sync tests
+  whenever possible, as sync tests run serially and async
+  tests run concurrently.
+
+  Finally, how many tests we have run, how many of them
+  failed, how many were invalid, and so on.
+
+  ### Understanding test failures
+
+  First, it contains the failure counter, followed by the test
+  name and the module the test was defined:
+
+      1) test greets the world (FooTest)
+
+  The next line contains the exact location of the test in the
+  `FILE:LINE` format:
+
+      test/foo_test.exs:5
+
+  If you want to re-run only this test, all you need to do is to
+  copy the line above and past it in front of `mix test`:
+
+      mix test test/foo_test.exs:5
+
+  Then we show the error message, code snippet, and general information
+  about the failed test:
+
+      Assertion with == failed
+      code:  assert Foo.hello() == :world!
+      left:  :world
+      right: :world!
+
+  If your terminal supports coloring (see the  "Coloring" section below),
+  a diff is typically shown between `left` and `right` sides. Finally,
+  we print the stacktrace of the failure:
+
+      stacktrace:
+        test/foo_test.exs:6: (test)
+
   ## Command line options
 
+    * `--all-warnings` - prints warnings even from files that do not need to be recompiled
+
     * `--color` - enables color in the output
+
     * `--cover` - runs coverage tool. See "Coverage" section below
+
     * `--exclude` - excludes tests that match the filter
-    * `--export-coverage` - the name of the file to export coverage results too.
+
+    * `--exit-status` - use an alternate exit status to use when the test suite
+      fails (default is 2).
+
+    * `--export-coverage` - the name of the file to export coverage results to.
       Only has an effect when used with `--cover`
+
     * `--failed` - runs only tests that failed the last time they ran
+
     * `--force` - forces compilation regardless of modification times
+
     * `--formatter` - sets the formatter module that will print the results.
       Defaults to ExUnit's built-in CLI formatter
+
     * `--include` - includes tests that match the filter
-    * `--listen-on-stdin` - runs tests, and then listens on stdin. Receiving a newline will
-      result in the tests being run again. Very useful when combined with `--stale` and
-      external commands which produce output on stdout upon file system modifications
+
+    * `--listen-on-stdin` - runs tests, and then listens on stdin. It will
+      re-run tests once a newline is received. See the "File system watchers"
+      section below
+
     * `--max-cases` - sets the maximum number of tests running asynchronously. Only tests from
       different modules run in parallel. Defaults to twice the number of cores
+
     * `--max-failures` - the suite stops evaluating tests when this number of test
       failures is reached. It runs all tests if omitted
+
     * `--no-archives-check` - does not check archives
+
     * `--no-color` - disables color in the output
+
     * `--no-compile` - does not compile, even if files require compilation
+
     * `--no-deps-check` - does not check dependencies
+
     * `--no-elixir-version-check` - does not check the Elixir version from `mix.exs`
+
     * `--no-start` - does not start applications after compilation
+
     * `--only` - runs only tests that match the filter
-    * `--partitions` - sets the amount of partitions to split tests in. This option
-      requires the `MIX_TEST_PARTITION` environment variable to be set. See the
-      "Operating system process partitioning" section for more information
+
+    * `--partitions` - sets the amount of partitions to split tests in. It must be
+      a number greater than zero. If set to one, it acts a no-op. If more than one,
+      then you must also set the `MIX_TEST_PARTITION` environment variable with the
+      partition to use in the current test run. See the "Operating system process
+      partitioning" section for more information
+
     * `--preload-modules` - preloads all modules defined in applications
+
+    * `--profile-require` - profiles the time spent to require test files.
+      Used only for debugging. The test suite does not run.
+
     * `--raise` - raises if the test suite failed
+
     * `--seed` - seeds the random number generator used to randomize the order of tests;
-      `--seed 0` disables randomization
+      `--seed 0` disables randomization so the tests in a single file will always be ran
+      in the same order they were defined in
+
     * `--slowest` - prints timing information for the N slowest tests.
       Automatically sets `--trace` and `--preload-modules`
+
     * `--stale` - runs only tests which reference modules that changed since the
       last time tests were ran with `--stale`. You can read more about this option
       in the "The --stale option" section below
+
     * `--timeout` - sets the timeout for the tests
+
     * `--trace` - runs tests with detailed reporting. Automatically sets `--max-cases` to `1`.
       Note that in trace mode test timeouts will be ignored as timeout is set to `:infinity`
 
+    * `--warnings-as-errors` - (since v1.12.0) treats warnings as errors and returns a non-zero
+      exit status. This option only applies to test files. To treat warnings as errors during
+      compilation and during tests, run:
+          MIX_ENV=test mix do compile --warnings-as-errors + test --warnings-as-errors
+
   ## Configuration
 
   These configurations can be set in the `def project` section of your `mix.exs`:
 
+    * `:test_coverage` - a set of options to be passed down to the coverage
+      mechanism. See the "Coverage" section for more information
+
+    * `:test_elixirc_options` - the compiler options to used when
+      loading/compiling test files. By default it disables the debug chunk
+      and docs chunk
+
     * `:test_paths` - list of paths containing test files. Defaults to
       `["test"]` if the `test` directory exists; otherwise, it defaults to `[]`.
       It is expected that all test paths contain a `test_helper.exs` file
@@ -86,8 +208,16 @@ defmodule Mix.Tasks.Test do
     * `:warn_test_pattern` - a pattern to match potentially misnamed test files
       and display a warning. Defaults to `*_test.ex`
 
-    * `:test_coverage` - a set of options to be passed down to the coverage
-      mechanism
+  ## Coloring
+
+  Coloring is enabled by default on most Unix terminals. They are also
+  available on Windows consoles from Windows 10, although it must be
+  explicitly enabled for the current user in the registry by running
+  the following command:
+
+      reg add HKCU\Console /v VirtualTerminalLevel /t REG_DWORD /d 1
+
+  After running the command above, you must restart your current console.
 
   ## Filters
 
@@ -146,26 +276,37 @@ defmodule Mix.Tasks.Test do
 
   ## Coverage
 
-  The `:test_coverage` configuration accepts the following options:
+  The `:test_coverage` configures the coverage tool and reports generated
+  by the `--cover` flag. It accepts the following options:
 
-    * `:output` - the output directory for cover results. Defaults to `"cover"`
-    * `:tool` - the coverage tool
-    * `:summary` - summary output configuration; can be either a boolean
-      or a keyword list. When a keyword list is passed, it can specify a
-      `:threshold`, which is a boolean or numeric value that enables coloring
-      of code coverage results in red or green depending on whether the
-      percentage is below or above the specified threshold, respectively.
-      Defaults to `[threshold: 90]`
-    * `:export` - a file name to export results to instead of generating
-      the result on the fly. The `.coverdata` extension is automatically
-      added to the given file. This option is automatically set via the
-      `--export-coverage` option or when using process partitioning.
+    * `:output` - the output directory for cover results. Defaults to `"cover"`.
+
+    * `:tool` - a module specifying the coverage tool to use.
+
+    * `:summary` - at the end of each coverage run, a summary of each
+      module is printed, with results in red or green depending on whether
+      the percentage is below or above a given threshold. The task will
+      exit with status of 1 if the total coverage is below the threshold.
+      The `:summary` option allows you to customize the summary generation
+      and defaults to `[threshold: 90]`, but it may be set to `false` to
+      disable such reports.
+
+    * `:export` - a filename to export results to instead of generating
+      the coverage result on the fly. The `.coverdata` extension is
+      automatically added to the given file. This option is automatically
+      set via the `--export-coverage` option or when using process partitioning.
       See `mix test.coverage` to compile a report from multiple exports.
-    * `:ignore_modules` - modules to ignore from generating reports and
-      in summaries
 
-  By default, a very simple wrapper around OTP's `cover` is used as a tool,
-  but it can be overridden as follows:
+    * `:ignore_modules` - modules to ignore from generating reports and
+      in summaries. It is a list of module names as atoms and regular
+      expressions that are matched against the module names.
+
+    * `:local_only` - by default coverage only tracks local calls,
+      set this option to false if you plan to run coverage across nodes.
+
+  By default, a wrapper around OTP's `cover` is used as the default coverage
+  tool. You can learn more about how it works in the docs for
+  `mix test.coverage`. Your tool of choice can be given as follows:
 
       def project() do
         [
@@ -177,8 +318,8 @@ defmodule Mix.Tasks.Test do
 
   `CoverModule` can be any module that exports `start/2`, receiving the
   compilation path and the `test_coverage` options as arguments.
-  It must return either `nil` or an anonymous function of zero arity that will
-  be run after the test suite is done.
+  It must return either `nil` or an anonymous function of zero arity that
+  will run after the test suite is done.
 
   ## Operating system process partitioning
 
@@ -226,9 +367,31 @@ defmodule Mix.Tasks.Test do
 
   The `--stale` option is extremely useful for software iteration, allowing you to
   run only the relevant tests as you perform changes to the codebase.
+
+  ## File-system watchers
+
+  You can integrate `mix test` with filesystem watchers through the command line
+  via the `--listen-on-stdin` option. For example, you can use [fswatch](https://github.com/emcrisostomo/fswatch)
+  or similar to emit newlines whenever there is a change, which will cause your test
+  suite to re-run:
+
+      fswatch lib test | mix test --listen-on-stdin
+
+  This can be combined with the `--stale` option to re-run only the test files that
+  have changed as well as the tests that have gone stale due to changes in `lib`.
+
+  ## Aborting the suite
+
+  It is possible to abort the test suite with `Ctrl+\ `, which sends a SIGQUIT
+  signal to the Erlang VM. ExUnit will intercept this signal to show all tests
+  that have been aborted and print the results collected so far.
+
+  This can be useful in case the suite gets stuck and you don't want to wait
+  until the timeout times passes (which defaults to 30 seconds).
   """
 
   @switches [
+    all_warnings: :boolean,
     force: :boolean,
     color: :boolean,
     cover: :boolean,
@@ -253,7 +416,10 @@ defmodule Mix.Tasks.Test do
     formatter: :keep,
     slowest: :integer,
     partitions: :integer,
-    preload_modules: :boolean
+    preload_modules: :boolean,
+    warnings_as_errors: :boolean,
+    profile_require: :string,
+    exit_status: :integer
   ]
 
   @cover [output: "cover", tool: Mix.Tasks.Test.Coverage]
@@ -266,7 +432,10 @@ defmodule Mix.Tasks.Test do
       do_run(opts, args, files)
     else
       {files_in_apps_path, files_not_in_apps_path} =
-        Enum.split_with(files, &String.starts_with?(&1, "apps/"))
+        files
+        |> Enum.map(&Path.expand/1)
+        |> Enum.map(&Path.relative_to_cwd/1)
+        |> Enum.split_with(&String.starts_with?(&1, "apps/"))
 
       app = Mix.Project.config()[:app]
       current_app_path = "apps/#{app}/"
@@ -316,17 +485,23 @@ defmodule Mix.Tasks.Test do
       """)
     end
 
-    # Load ExUnit before we compile anything
+    # Load ExUnit before we compile anything in case we are compiling
+    # helper modules that depend on ExUnit.
     Application.ensure_loaded(:ex_unit)
 
-    Mix.Task.run("compile", args)
+    # --warnings-as-errors in test does not pass down to compile,
+    # if you need this, call compile explicitly before.
+    Mix.Task.run("compile", args -- ["--warnings-as-errors"])
+
     project = Mix.Project.config()
+    {partitions, opts} = Keyword.pop(opts, :partitions)
+    partitioned? = is_integer(partitions) and partitions > 1
 
     # Start cover after we load deps but before we start the app.
     cover =
       if opts[:cover] do
         compile_path = Mix.Project.compile_path(project)
-        partition = opts[:partitions] && System.get_env("MIX_TEST_PARTITION")
+        partition = partitioned? && System.get_env("MIX_TEST_PARTITION")
 
         cover =
           @cover
@@ -355,6 +530,7 @@ defmodule Mix.Tasks.Test do
     ExUnit.configure(merge_helper_opts(ex_unit_opts))
 
     # Finally parse, require and load the files
+    test_elixirc_options = project[:test_elixirc_options] || []
     test_files = parse_files(files, shell, test_paths)
     test_pattern = project[:test_pattern] || "*_test.exs"
     warn_test_pattern = project[:warn_test_pattern] || "*_test.ex"
@@ -363,11 +539,11 @@ defmodule Mix.Tasks.Test do
       test_files
       |> Mix.Utils.extract_files(test_pattern)
       |> filter_to_allowed_files(allowed_files)
-      |> filter_by_partition(shell, opts)
+      |> filter_by_partition(shell, partitions)
 
     display_warn_test_pattern(test_files, test_pattern, matched_test_files, warn_test_pattern)
 
-    case CT.require_and_run(matched_test_files, test_paths, opts) do
+    case CT.require_and_run(matched_test_files, test_paths, test_elixirc_options, opts) do
       {:ok, %{excluded: excluded, failures: failures, total: total}} ->
         Mix.shell(shell)
         cover && cover.()
@@ -377,7 +553,9 @@ defmodule Mix.Tasks.Test do
             raise_with_shell(shell, "\"mix test\" failed")
 
           failures > 0 ->
-            System.at_exit(fn _ -> exit({:shutdown, 1}) end)
+            System.at_exit(fn _ ->
+              exit({:shutdown, Keyword.fetch!(ex_unit_opts, :exit_status)})
+            end)
 
           excluded == total and Keyword.has_key?(opts, :only) ->
             message = "The --only option was given to \"mix test\" but no test was executed"
@@ -446,7 +624,8 @@ defmodule Mix.Tasks.Test do
     :slowest,
     :failures_manifest_file,
     :only_test_ids,
-    :test_location_relative_path
+    :test_location_relative_path,
+    :exit_status
   ]
 
   @doc false
@@ -460,6 +639,7 @@ defmodule Mix.Tasks.Test do
       |> filter_opts(:only)
       |> formatter_opts()
       |> color_opts()
+      |> exit_status_opts()
       |> Keyword.take(@option_keys)
       |> default_opts()
 
@@ -563,34 +743,40 @@ defmodule Mix.Tasks.Test do
     Enum.filter(matched_test_files, &MapSet.member?(allowed_files, Path.expand(&1)))
   end
 
-  defp filter_by_partition(files, shell, opts) do
-    if total = opts[:partitions] do
-      partition = System.get_env("MIX_TEST_PARTITION")
+  defp filter_by_partition(files, _shell, total) when total in [nil, 1],
+    do: files
 
-      case partition && Integer.parse(partition) do
-        {partition, ""} when partition in 1..total ->
-          partition = partition - 1
+  defp filter_by_partition(files, shell, total) when total > 1 do
+    partition = System.get_env("MIX_TEST_PARTITION")
 
-          # We sort the files because Path.wildcard does not guarantee
-          # ordering, so different OSes could return a different order,
-          # meaning run across OSes on different partitions could run
-          # duplicate files.
-          for {file, index} <- Enum.with_index(Enum.sort(files)),
-              rem(index, total) == partition,
-              do: file
+    case partition && Integer.parse(partition) do
+      {partition, ""} when partition in 1..total ->
+        partition = partition - 1
 
-        _ ->
-          raise_with_shell(
-            shell,
-            "The MIX_TEST_PARTITION environment variable must be set to an integer between " <>
-              "1..#{total} when the --partitions option is set, got: #{inspect(partition)}"
-          )
-      end
-    else
-      files
+        # We sort the files because Path.wildcard does not guarantee
+        # ordering, so different OSes could return a different order,
+        # meaning run across OSes on different partitions could run
+        # duplicate files.
+        for {file, index} <- Enum.with_index(Enum.sort(files)),
+            rem(index, total) == partition,
+            do: file
+
+      _ ->
+        raise_with_shell(
+          shell,
+          "The MIX_TEST_PARTITION environment variable must be set to an integer between " <>
+            "1..#{total} when the --partitions option is set, got: #{inspect(partition)}"
+        )
     end
   end
 
+  defp filter_by_partition(_files, shell, total) do
+    raise_with_shell(
+      shell,
+      "--partitions : expected to be positive integer, got #{total}"
+    )
+  end
+
   defp color_opts(opts) do
     case Keyword.fetch(opts, :color) do
       {:ok, enabled?} ->
@@ -601,6 +787,10 @@ defmodule Mix.Tasks.Test do
     end
   end
 
+  defp exit_status_opts(opts) do
+    Keyword.put_new(opts, :exit_status, 2)
+  end
+
   defp require_test_helper(shell, dir) do
     file = Path.join(dir, "test_helper.exs")
 
